/* Tests that the hash code is correctly computed when running concurrently.
   Regression test for invalid use of static variables in Standard_GUID::Hash().
   and constructor Standard_GUID (const Standard_ExtString).
*/

struct HashData {
    TCollection_AsciiString myString;
    Standard_GUID       myGuid;
    Standard_Integer    myHash;
};

class ConcurrentHashFunctor
{
public:

    static Standard_Integer Upper() { return 101; }

    ConcurrentHashFunctor (const HashData* theExpected, const size_t theLength,
        bool* theFailFlag) :
        myArr (theExpected), myLength (theLength), myFailure (theFailFlag) {}

    void operator() (const tbb::blocked_range<size_t>& r) const
    {
        //return if at least one thread reported failure
        if (*myFailure) return;

        //compute the index for this thread in myArr
        size_t anIndex = OSD_Thread::Current() % myLength;
        const Standard_GUID& aGuid = myArr[anIndex].myGuid;
        Standard_Integer anExpectedHash = myArr[anIndex].myHash;

        //now re-compute and compare
        for (size_t i = r.begin(); i != r.end() && !*myFailure; ++i) {
            //1. tests Standard_GUID::Hash()
            Standard_Integer aHash = aGuid.Hash (Upper());
            if (aHash != anExpectedHash) {
                *myFailure = true;
                return;
            }
            //2. test constructor Standard_GUID (const Standard_ExtString)
            TCollection_ExtendedString anExtendedString (myArr[anIndex].myString);
            Standard_GUID aGuid2 (anExtendedString.ToExtString());
            aHash = aGuid2.Hash (Upper());
            if (aHash != anExpectedHash) {
                *myFailure = true;
                return;
            }
        }
    }
private:

    const HashData* myArr;
    size_t          myLength;
    bool*           myFailure;

};

void Standard_GUIDTest::Concurrent()
{
    const size_t aNb = 11; //should be some prime number

    //sample GUIDs
    char* aGuidArr[aNb] = {
        "ea55eea6-4e9f-47e5-8a07-5cd11099b310",
        "48f415c7-7aa3-47be-9f53-c69178fa27d2",
        "018c5486-99a2-4418-9ab1-93dc4d9e0c67",
        "b944cc19-1f50-436a-a0f5-604a99f229e9",
        "d06f75a8-8829-4e19-a207-ff54cb458117",
        "7b74f6cf-a617-4319-955b-a8fd71e14885",
        "2b8a5c19-e679-47eb-9327-dbe489c0c4b7",
        "e612fbb1-fd87-4643-9ddc-67c47e78a987",
        "53b647cb-4c36-4ca6-84d9-b580c5439ea2",
        "9cf5c67f-f3fb-484d-9837-207a81b54e0b",
        "be8481ae-5fb6-4c3c-b792-48f037a4308a"
    };

    //compute expected hashes in single thread
    HashData anArr[aNb];
    for (size_t i = 0; i < aNb; ++i) {
        anArr[i].myString = aGuidArr[i];
        anArr[i].myGuid = aGuidArr[i];
        anArr[i].myHash = anArr[i].myGuid.Hash (ConcurrentHashFunctor::Upper());
    }

    //now run concurrently and re-compute
    bool aFailure = false;
    const int N = 1 << 20; //big enough to ensure concurrency

    tbb::parallel_for (tbb::blocked_range<size_t> (0, N),
        ConcurrentHashFunctor (anArr, aNb, &aFailure));
    QVERIFY (!aFailure);
}
